/* just some information about this file */
#define DOUBLELINKEDLIST_VERSION	"1.1.1"
#define DOUBLELINKEDLIST_DATE	"04-Dec-2003"

/* structures and typedefs */

/** @file
* Fault-tolerant double linked list data structure.
*
* This structure provides a fault-tolerant data structure of a double linked list. Basically, a double
* linked list is a dynamic data structure. This implementation facilitates insertions and removals of
* nodes. A queue and a stack are emulated with 'define' functions (see below).
*
* Two structures are used to implement the list:
* @li @ref listNode (self-referential structure)
* @li @ref List
* <BR>
* represent a node of the list and the complete double linked list, respectively. As the data of
* each node is just a @c void* every kind of data can be stored.
*
* The structure is created with @ref List_create, destroyed with @ref List_destroy and
* displayed with @ref List_display. The list is traversed using @ref List_getNext and
* @ref List_hasNext. Elements are added at the current position with
* @ref List_insert and removed with @ref List_remove. Additionally, a node is added at the end of the
* list with @ref List_insertTail and at the beginning with @ref List_insertHead. Nodes are removed from
* the head of the list with @ref List_removeHead and from the end of the list with @ref List_removeTail.
* As the list keeps track of the current node during traversal (changed with @ref List_getNext) it is
* rewinded with @ref List_rewind. The function @ref List_isEmpty returns @ref BOOLEAN_TRUE if the list
* is empty and @ref BOOLEAN_FALSE otherwise.
*
* The structure @ref List provides also a possibility to store a name for this list as a char*. The
* name can be accessed with @ref List_setName and @ref List_getName. The structure keeps automatically
* track of the current number of nodes. This value is returned by @ref List_getNumberOfNodes.
* <BR>
* Stacks are implemented by the define @c functions @ref List_pushhead, @ref List_pophead and @ref
* List_top. Functionality of a queue is provided by @ref List_enqueue and @ref List_dequeue.
* <BR><BR>
* One has to provide a function pointer to display the data that is stored in the individual nodes via
* the @c void*! The display function has to comply to the prototype <b>void
* displayFunction( void*, FILE* )</b>. A function pointer to destroy (i.e. free) the data of a node is
* also mandatory. The destroy function has to comply to the prototype <b>void destroyFunction
* ( void* )</b>.
* <BR><BR>
*	The list can automatically check for every insertion operation if there exists already a node with a
* void* @c data that is identical to the one that should be inserted. This has to be specified with a
* function parameter during creation of the List (@ref List_create) and cannot be changed afterwards.
* However, function @ref List_getUniqueData can be used to query a list about its 'uniqueness' status.
* The status is queried with @ref List_getUniqueData. Uniqueness is guaranteed with the function
* @c identicalFunction; it is provided as a function parameter to @ref List_create. The provided
* function @c identicalFunction has to return BOOLEAN_FALSE or BOOLEAN_TRUE if the data is identical or
* non-identical, respectively.
* <P>
* @anchor listNode
* <B>Structure listNode</B>
* <BR><BR>
* The structure listNode consists of a @c void* to store the data, of a pointer to the next node in
* the list and of a pointer to the previous node (self-referential structure).
*
* <B>Description of the members of listNode:</B><BR><BR>
* @c void* - points to the data of the node<BR>
* @c struct listNode* - pointer on the <b>previous</b> node in the list<BR>
* @c struct listNode* - pointer on the <b>next</b> node in the list<BR>
* <P>
*
* @anchor List
* <B>Structure List</B>
* <BR><BR>
* The structure List conists of a ListNode_Ptr on the first node, a ListNode_Ptr on the
* last node, a pointer on the current node (needed for sequential traversal of the list) and a
* function pointer to display the data of a node. Additionally, a name (char*) and the current number
* of nodes is stored and accessible with appropriate functions.
*
* <B>Description of the members of SingleLinkedList:</B><BR><BR>
* @c name - char* to store the name of the list<BR>
* @c numberOfNodes - stores the number of nodes in the list<BR>
* @c headNodePtr - ListNode_Ptr on the first node in the list<BR>
* @c tailNodePtr - ListNode_Ptr on the last node in the list<BR>
* @c currentNodePtr - ListNode_Ptr on the current node in the list<BR>
* @c uniqueData - indicates via BOOLEAN_TRUE and BOOLEAN_FALSE if nodes with identical data are allowed<BR>
* @c displayFunction - pointer to a function that is capable of displaying the data of a node<BR>
* @c destroyFunction - pointer to a function that is capable of destroying the data of a node<BR>
* @c identicalFunction - pointer to a function that returns whether the data of two nodes is identical<BR>
* <P>
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
* @version 03/12/2003 - Uli Fechner - added support for lists that contain only non-identical @c data
* in their nodes (added member @c uniqueData and @c identicalFunction to structure List; added functions
* @ref List_isContained, @ref List_getUniqueData; modified @ref List_insert, @ref List_insertHead and
* @ref List_insertTail); the currentNodePtr
* @version 04/12/2003 - Uli Fechner - changed the behaviour of re-setting @c currentNodePtr in @ref
* List_remove and @ref List_removeHead for consistency reasons
* @@code
*/
struct listNode
{
	void* data;
	struct listNode* previousNodePtr;
	struct listNode* nextNodePtr;
};

/** A structure @ref listNode is assigned the name @c ListNode. */
typedef struct listNode ListNode;

/** A pointer to structure @ref ListNode is assigned the name @c ListNode_Ptr. */
typedef ListNode* ListNode_Ptr;

typedef struct
{
	char* name;
	int numberOfNodes;
	ListNode_Ptr headNodePtr;
	ListNode_Ptr tailNodePtr;
	ListNode_Ptr currentNodePtr;
	int uniqueData;
	void (*displayFunction) ( void* data, FILE* outputStream );
	void (*destroyFunction) ( void* data );
	int (*identicalFunction) ( const void* data1, const void* data2 );
} List;

/** A pointer to structure @ref List is assigned the name @c List_Ptr. */
typedef List* List_Ptr;
/** @endcode */

/* defines */

/** If DOUBLELINKESLIST_DEBUG is set to 1 some debug information is written to standard error.
*
* @author Uli Fechner
* @version 25/11/2003 - Uli Fechner - initial release
*/
#ifndef DOUBLELINKESLIST_DEBUG
#define DOUBLELINKEDLIST_DEBUG 0
#endif

/** Boolean variables are simulated with the defines BOOLEAN_FALSE and BOOLEAN_TRUE.
*
* @author Uli Fechner
* @version 03/12/2003 - Uli Fechner - initial release
*/
#ifndef BOOLEAN_FALSE
#define BOOLEAN_FALSE 0
#endif

/** Boolean variables are simulated with the defines BOOLEAN_FALSE and BOOLEAN_TRUE.
*
* @author Uli Fechner
* @version 03/12/2003 - Uli Fechner - initial release
*/
#ifndef BOOLEAN_TRUE
#define BOOLEAN_TRUE 1
#endif

/* define 'functions' */
/* The following defines are used to emulate function calls for stacks */

/** Pushes an element on top of a <b>stack</b>.
* 
* @attention
* This define is only a wrapper for @ref List_insertHead!
*
* @author Uli Fechner
* @version 25/11/2003 - Uli Fechner - initial release
*/
#define List_pushhead( listPtr, data ) \
	( List_insertHead( listPtr, data ) )

/** Pops an element from top of a <b>stack</b>.
* 
* @attention
* This define is only a wrapper for @ref List_removeHead!
*
* @author Uli Fechner
* @version 25/11/2003 - Uli Fechner - initial release
*/
#define List_pophead( listPtr ) \
	( List_removeHead( listPtr ) )

/** Returns the element on top of a <b>stack</b> without removing it.
* 
* @attention
* This define wraps sequentially @ref List_rewind and @ref List_getNext!
*
* @author Uli Fechner
* @version 25/11/2003 - Uli Fechner - initial release
*/
#define List_top( listPtr ) \
	( ( List_rewind( listPtr ) ), \
	( List_getNext( listPtr ) ) )

/* The following defines are used to emulate function calls for queues */

/** Enqueues an element in a <b>queue</b>.
* 
* @attention
* This define is only a wrapper for @ref List_insertTail!
*
* @author Uli Fechner
* @version 25/11/2003 - Uli Fechner - initial release
*/
#define List_enqueue( listPtr, data ) \
	( List_insertTail( listPtr, data ) )

/** Dequeues an element of a <b>queue</b>.
* 
* @attention
* This define is only a wrapper for @ref List_removeTail!
*
* @author Uli Fechner
* @version 25/11/2003 - Uli Fechner - initial release
*/
#define List_dequeue( listPtr ) \
	( List_removeHead( listPtr ) )

/* function prototypes */

List_Ptr List_create( const int uniqueData, void (*displayFunction) ( void* data, FILE* outputStream ), \
	void (*destroyFunction) ( void* data ), \
	int (*compareFunction) ( const void* data1, const void* data2 ) );

void List_destroy( List_Ptr listPtr );

void List_display( const List_Ptr listPtr, FILE* outputStream );

void* List_getNext( const List_Ptr listPtr );

int List_hasNext( const List_Ptr listPtr );

void List_rewind( const List_Ptr listPtr );

int List_insert( const List_Ptr listPtr, void* data );

void* List_remove( const List_Ptr listPtr );

int List_insertHead( const List_Ptr listPtr, void* data );

int List_insertTail( const List_Ptr listPtr, void* data );

void* List_removeHead( const List_Ptr listPtr );

void* List_removeTail( const List_Ptr listPtr );

int List_isEmpty( const List_Ptr listPtr );

int List_isContained( const List_Ptr listPtr, const void* data );

int List_getNumberOfNodes( const List_Ptr listPtr );

void List_setName( const List_Ptr listPtr, char* name );

char* List_getName( const List_Ptr listPtr );

int List_getUniqueData( const List_Ptr listPtr );

/* functions */

/** Creates an @ref List structure.
*
* A structure @ref List is created. The memory of the structure itself and of its members is
* allocated automatically. A function pointer on a function to display the data that is stored in this
* List has to provided as an argument.
*
* @param uniqueData indicates via a boolean variable whether the data of the list should be checked for uniqueness
* @param displayFunction function pointer on a function that is capable of displaying the data of a node
* @param destroyFunction function pointer on a function that is capable of destroying the data of a node
* @param identicalFunction function pointer on a function that is capable of checking whether the data of two nodes is identical
* @retval List_Ptr pointer on the newly created structure @ref List
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
* @version 03/12/2003 - Uli Fechner - added function parameters uniqueData and compareFunction
*/
List_Ptr List_create( const int uniqueData, void (*displayFunction) ( void* data, FILE* outputStream ), \
	void (*destroyFunction) ( void* data ), \
	int (*identicalFunction) ( const void* data1, const void* data2 ) )
{
	List_Ptr listPtr;	
	if( !( listPtr = calloc( 1, sizeof( List ) ) ) )
		MemoryError( "listPtr", "List_create" );
	
	if( displayFunction == NULL )
	{
		fprintf( stderr, "\nERROR: The argument (*displayFunction) of the function\n" );
		fprintf( stderr, "List_create must not be 'NULL'!\n" );
		AbortProgram;
	}
	
	if( destroyFunction == NULL )
	{
		fprintf( stderr, "\nERROR: The argument (*destroyFunction) of the function\n" );
		fprintf( stderr, "List_create must not be 'NULL'!\n" );
		AbortProgram;
	}

	listPtr->name = NULL;
	listPtr->numberOfNodes = 0;
	listPtr->headNodePtr = NULL;
	listPtr->tailNodePtr = NULL;
	listPtr->currentNodePtr = NULL;
	listPtr->uniqueData = uniqueData;
	listPtr->displayFunction = displayFunction;
	listPtr->destroyFunction = destroyFunction;
	listPtr->identicalFunction = identicalFunction;
	
	return listPtr;
}

/** Destroys a @ref List structure.
*
* The structure @ref List the pointer @c listPtr refers to is destroyed. The memory of the
* structure itself and of its members is freed automatically.
*
* @attention
* The data of all nodes is freed with a trivial free of the node void pointers! If the void pointer
* refers to structures that have members that allocate memory this memory is not freed! In such a case
* it is best to remove the nodes one by one and free the memory individually.
*
* @param listPtr pointer on structure @ref List that should be destroyed
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
*/
void List_destroy( List_Ptr listPtr )
{
	void* data;
	
	if( listPtr == NULL )
	{
		fprintf( stderr, "\n\nERROR: Function 'List_destroy' can not destroy a structure\n" );
		fprintf( stderr, "'List' that has not been created before!\n" );
		AbortProgram;
	}

	List_rewind( listPtr ); // re-set the currentNodePtr
	while( !List_isEmpty( listPtr ) )
	{
		data = List_removeHead( listPtr );
		listPtr->destroyFunction( data );
	}	
	/* destroying the struct List itself */
	free( listPtr );
}

/** Displays a structure @ref List.
*
* The structure @ref List the pointer @c listPtr refers to is displayed on the FILE* @c
* outputStream. If the list is empty an appropriate message is printed on @c outputStream.
*
* @param listPtr pointer on the structure @ref List that should be displayed
* @param outputStream FILE* on the stream the output should be sent to
* @author Uli Fechner
* @version 25/11/2003 - Uli Fechner - initial release
* @version 03/12/2003 - Uli Fechner - added support for the members @c uniqueData and identicalFunction
*/
void List_display( const List_Ptr listPtr, FILE* outputStream )
{
	ListNode_Ptr tempNodePtr;	
	tempNodePtr = listPtr->headNodePtr;
	
	#if DOUBLELINKEDLIST_DEBUG
		fprintf( stderr, "list %p\nname: %s\nnumberOfNodes: %d\nuniqueData: %d\n", listPtr, listPtr->name, \
			listPtr->numberOfNodes, listPtr->uniqueData );
		fprintf( stderr, "headNodePtr: %p\ntailNodePtr: %p\ncurrentNodePtr: %p\n", listPtr->headNodePtr, \
			listPtr->tailNodePtr, listPtr->currentNodePtr );
		fprintf( stderr, "displayFunction: %p\ndestroyFunction: %p\nidenticalFunction: %p\n", \
			listPtr->displayFunction, listPtr->destroyFunction, listPtr->identicalFunction );
	#endif		
	/* if the list is empty an appropriate message is printed on outputStream */
	if( tempNodePtr == NULL )
		fprintf( outputStream, "The list is empty.\n" );
	/* if the list is not empty we go from node to node and use displayFunction to output the data */
	while( tempNodePtr != NULL )
	{
		#if DOUBLELINKEDLIST_DEBUG
			fprintf( stderr, "current node: %p\n", tempNodePtr );
			fprintf( stderr, "previous node: %p\n", tempNodePtr->previousNodePtr );
			fprintf( stderr, "next node: %p\n", tempNodePtr->nextNodePtr );
		#endif		
		listPtr->displayFunction( tempNodePtr->data, outputStream );
		tempNodePtr = tempNodePtr->nextNodePtr;
	}
	fprintf( outputStream, "\n" );
}

/** Returns the data of the next node in the list.
*
* The current position within the @ref List is stored in the member @c currentNodePtr.
*
* @param listPtr pointer on the structure @ref List the node belongs to
* @retval void* on the data of the node
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
*/
void* List_getNext( const List_Ptr listPtr )
{
	/* checking if the currentNodePtr points 'before' the first node */
	if( listPtr->currentNodePtr == NULL )
	{
		/* checking if the list is empty */
		if( !List_isEmpty( listPtr ) )
		{
			listPtr->currentNodePtr = listPtr->headNodePtr;
			return (listPtr->currentNodePtr)->data; // FUNCTION EXITS HERE
		}
		/* list is empty */
		else
		{
			fprintf( stderr, "\nERROR: The double linked list [%p] is empty!\n", listPtr );
			fprintf( stderr, "Function List_getNext must not be called for an empty list\n!" );
			AbortProgram;
		}
	}

	/* checking if currentNodePtr points already to the last node in the list */
	if( listPtr->currentNodePtr == listPtr->tailNodePtr )
	{
		fprintf( stderr, "\nERROR: Function List_getNext reached the end of the double linked\n" );
		fprintf( stderr, "list [%s; %p]! Boundaries of the single linked list must be checked!\n", \
			List_getName( listPtr ), listPtr );
		AbortProgram;
	}
	
	/* setting currentNodePtr to the next node */
	listPtr->currentNodePtr = (listPtr->currentNodePtr)->nextNodePtr;
	
	return (listPtr->currentNodePtr)->data;
}

/** Checks if the @ref List has at least one more node (directed forwards).
*
* The current position within the @ref List is stored in the member @c currentNodePtr. This function
* checks if @c currentNodePtr points to @c tailNodePtr. If so, the end of the list is reached.
*
* @param listPtr pointer on the structure @ref List that is checked
* @retval int BOOLEAN_FALSE or BOOLEAN_TRUE
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
*/
int List_hasNext( const List_Ptr listPtr )
{
	if( listPtr->currentNodePtr == listPtr->tailNodePtr )
		return BOOLEAN_FALSE;
	else
		return BOOLEAN_TRUE;
}

/** Re-sets the position pointer of the @ref List to the beginning of the list.
*
* The current position within the @ref List is stored in the member @c currentNodePtr. This function
* re-sets @c currentNodePtr to the very beginning of the list, so that the next call to @ref
* List_getNext returns the data of the first node.
*
* @param listPtr pointer on the structure @ref List the position pointer is re-set
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
*/
void List_rewind( const List_Ptr listPtr )
{
	listPtr->currentNodePtr = NULL;
}

/** A new node is inserted in @ref List at the position that is referred by @c currentNodePtr.
*
*	@attention
* After a successful insertion the @c currentNodePtr points to the newly created node.
* 
* @param listPtr pointer on the structure @ref List the node is inserted
* @param data data of the new node
* @retval int returns BOOLEAN_TRUE on a successful and BOOLEAN_FALSE on a unsuccessful insertion
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
* @version 03/12/2003 - Uli Fechner - changed return type from void to int (BOOLEAN_TRUE and
* BOOLEAN_FALSE); added support for uniqueData list
*/
int List_insert( const List_Ptr listPtr, void* data )
{
	ListNode_Ptr newListNodePtr;
	int successfulInsertion;
	
	/* node is inserted as the first one in the list */
	if( listPtr->currentNodePtr == NULL )
	{
		successfulInsertion = List_insertHead( listPtr, data );
		listPtr->currentNodePtr = listPtr->headNodePtr;
		return successfulInsertion;
	}
	
	/* node is inserted as the last one in the list */
	if( listPtr->currentNodePtr == listPtr->tailNodePtr )
	{
		successfulInsertion = List_insertTail( listPtr, data );
		listPtr->currentNodePtr = listPtr->tailNodePtr;
		return successfulInsertion;
	}
		
	/* if the list allows only unique data we check here if the data to be inserted is already in the
	list; if we find a duplicate NULL is returned */
	if( listPtr->uniqueData == BOOLEAN_TRUE )
	{
		if( ( List_isContained( listPtr, data ) ) == BOOLEAN_TRUE )
			return BOOLEAN_FALSE;
	}
	
	if( !( newListNodePtr = calloc( 1, sizeof( ListNode ) ) ) )
		MemoryError( "newListNodePtr", "List_insert" );
	newListNodePtr->data = data;
	newListNodePtr->nextNodePtr = NULL;
	newListNodePtr->previousNodePtr = NULL;
	
	newListNodePtr->previousNodePtr = listPtr->currentNodePtr;
	newListNodePtr->nextNodePtr = (listPtr->currentNodePtr)->nextNodePtr;
	((listPtr->currentNodePtr)->nextNodePtr)->previousNodePtr = newListNodePtr;
	(listPtr->currentNodePtr)->nextNodePtr = newListNodePtr;
	
	listPtr->currentNodePtr = newListNodePtr;
	(listPtr->numberOfNodes)++;
	
	return BOOLEAN_TRUE;
}

/** The node that is referred to by @c currentNodePtr is removed from @ref List and its data is returned.
*
*	@attention
* The @c currentNodePtr is moved to the previous node. If the head node is removed @c currentNodePtr is
* moved <b>before</b> the new head node. This behaviour guarantees that the following call of @ref
* List_getNext returns the element that follows the one that was removed.
* 
* @param listPtr pointer on the structure @ref List the node is removed from
* @retval void* data of the removed node
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
* @version 04/12/2003 - Uli Fechner - changed the behaviour of @c currentNodePtr
*/
void* List_remove( const List_Ptr listPtr )
{
	void* data;
	/* ListNode_Ptr on the previous and the next node of listPtr->currentNodePtr */
	ListNode_Ptr previousNodePtr, nextNodePtr;
	
	/* returns with a function call to List_removeHead if currentNodePtr points to headNodePtr */
	if( listPtr->currentNodePtr == listPtr->headNodePtr )
		return List_removeHead( listPtr );
	
	/* returns with a function call to List_removeTail if currentNodePtr points to tailNodePtr */
	if( listPtr->currentNodePtr == listPtr->tailNodePtr )
		return List_removeTail( listPtr );
	
	if( listPtr->currentNodePtr == NULL )
	{
		fprintf( stderr, "\nERROR: currentNodePtr points before the first node of the\n" );
		fprintf( stderr, "list [%s; %p]. Removal is not possible.\n", List_getName( listPtr ), listPtr );
		AbortProgram;
	}
	
	previousNodePtr = (listPtr->currentNodePtr)->previousNodePtr;
	nextNodePtr = (listPtr->currentNodePtr)->nextNodePtr;
	previousNodePtr->nextNodePtr = nextNodePtr;
	nextNodePtr->previousNodePtr = previousNodePtr;
	data = (listPtr->currentNodePtr)->data;
	free( listPtr->currentNodePtr );
	listPtr->currentNodePtr = previousNodePtr;
	//listPtr->currentNodePtr = nextNodePtr;
	(listPtr->numberOfNodes)--;

	return data;
}

/** Inserts a node at the head of the @ref List.
*
* @attention
* The @c currentNodePtr is not changed if this function is called directly.
*
* @param listPtr pointer on the structure @ref List the new node is added
* @param data void pointer on the actual data
* @retval int returns BOOLEAN_TRUE on a successful and BOOLEAN_FALSE on an unsuccessful insertion
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
* @version 03/12/2003 - Uli Fechner - changed return type from void to int (BOOLEAN_TRUE and
* BOOLEAN_FALSE); added support for uniqueData list
*/
int List_insertHead( const List_Ptr listPtr, void* data )
{
	ListNode_Ptr newListNodePtr;

	/* if the list allows only unique data we check here if the data to be inserted is already in the
	list; if we find a duplicate NULL is returned */
	if( listPtr->uniqueData == BOOLEAN_TRUE )
	{
		if( ( List_isContained( listPtr, data ) ) == BOOLEAN_TRUE )
			return BOOLEAN_FALSE;
	}

	if( !( newListNodePtr = calloc( 1, sizeof( ListNode ) ) ) )
		MemoryError( "newListNodePtr", "List_insertHead" );
	newListNodePtr->data = data;
	newListNodePtr->nextNodePtr = NULL;
	newListNodePtr->previousNodePtr = NULL;
	
	/* sets the currentNodePtr to the inserted node */
	//listPtr->currentNodePtr = newListNodePtr;
	
	/* if the list is empty tailNodePtr has to point on the newly added node */
	if( listPtr->headNodePtr == NULL )
	{
		listPtr->tailNodePtr = newListNodePtr;
	}
	/* if the list is non-empty the previousNodePtr of the first node has to point on the newly added node
	and the nextNodePtr of the newly added node has to point on headNodePtr */
	else
	{
		(listPtr->headNodePtr)->previousNodePtr = newListNodePtr;
		newListNodePtr->nextNodePtr = listPtr->headNodePtr;
	}
	
	listPtr->headNodePtr = newListNodePtr;
	(listPtr->numberOfNodes)++;
	
	return BOOLEAN_TRUE;
}

/** Inserts a node at the tail of the @ref List.
*
* @attention
* The @c currentNodePtr is not changed if this function is called directly.
*
* @param listPtr pointer on the structure @ref List the new node is added
* @param data void pointer on the actual data
* @retval int returns BOOLEAN_TRUE on a successful and BOOLEAN_FALSE on an unsuccessful insertion
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
* @version 03/12/2003 - Uli Fechner - changed return type from void to int (BOOLEAN_TRUE and
* BOOLEAN_FALSE); added support for uniqueData list
*/
int List_insertTail( const List_Ptr listPtr, void* data )
{
	ListNode_Ptr newListNodePtr;

	/* if the list allows only unique data we check here if the data to be inserted is already in the
	list; if we find a duplicate NULL is returned */
	if( listPtr->uniqueData == BOOLEAN_TRUE )
	{
		if( ( List_isContained( listPtr, data ) ) == BOOLEAN_TRUE )
			return BOOLEAN_FALSE;
	}
	
	if( !( newListNodePtr = calloc( 1, sizeof( ListNode ) ) ) )
		MemoryError( "newListNodePtr", "List_insertTail" );
	newListNodePtr->data = data;
	newListNodePtr->nextNodePtr = NULL;
	newListNodePtr->previousNodePtr = NULL;
	
	/* sets the currentNodePtr to the inserted node */
	//listPtr->currentNodePtr = newListNodePtr;

	/* if the list is empty headNodePtr has to point on the newly added node */
	if( listPtr->headNodePtr == NULL )
	{
		listPtr->headNodePtr = newListNodePtr;
	}
	/* if the list is non-empty the nextNodePtr of the last node has to point on the newly added node
	and the previousNodePtr of the newly added node has to point on tailNodePtr */
	else
	{
		(listPtr->tailNodePtr)->nextNodePtr = newListNodePtr;
		newListNodePtr->previousNodePtr = listPtr->tailNodePtr;
	}
	
	listPtr->tailNodePtr = newListNodePtr;
	(listPtr->numberOfNodes)++;
	
	return BOOLEAN_TRUE;
}

/** The first node of the structure @ref List is removed.
*
*	@attention
* If the @c currentNodePtr points to the head node it is moved before the new head node. Otherwise it is
* left untouched. This behaviour guarantees that a call to @ref List_getNext retrieves the node that
* follows the one that was removed.
* 
* @param listPtr pointer on the structure @ref List the node is removed from
* @retval void* data of the removed node
* @author Uli Fechner
* @version 28/11/2003 - Uli Fechner - initial release
* @version 04/12/2003 - Uli Fechner - changed the behaviour of the currentNodePtr
*/
void* List_removeHead( const List_Ptr listPtr )
{
	void* data;
	ListNode_Ptr newHeadNodePtr; /* temporary storage for the new head node pointer of list */
	
	/* nothing can be removed if the list is empty */
	if( listPtr->headNodePtr == NULL )
	{
		fprintf( stderr, "\nERROR: Structure List [%p] is empty!\n", listPtr );
		fprintf( stderr, "Function 'List_removeHead' must not be called for an empty list.\n" );
		AbortProgram;
	}
	
	/* if currentNodePtr points to the head node it is moved to the next node */
	if( listPtr->currentNodePtr == listPtr->headNodePtr )
		listPtr->currentNodePtr = NULL;
		//listPtr->currentNodePtr = (listPtr->currentNodePtr)->nextNodePtr;

	data = (listPtr->headNodePtr)->data;
	newHeadNodePtr = (listPtr->headNodePtr)->nextNodePtr;
	if( newHeadNodePtr != NULL )
		newHeadNodePtr->previousNodePtr = NULL;
	free( listPtr->headNodePtr );
	listPtr->headNodePtr = newHeadNodePtr;

	/* if the last node is removed from the list tailNodePtr is set to NULL */
	if( listPtr->headNodePtr == NULL )
		listPtr->tailNodePtr = NULL;
	
	(listPtr->numberOfNodes)--;
	return data;
}

/** The last node of the structure @ref List is removed.
*
*	@attention
* If the @c currentNodePtr points to the tail node it is moved to the new tail node. Otherwise it is
* left untouched.
*
* @param listPtr pointer on the structure @ref List the node is removed from
* @retval void* data of the removed node
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
*/
void* List_removeTail( const List_Ptr listPtr )
{
	void* data;
	ListNode_Ptr newTailNodePtr; /* temporary storage for the new tail node pointer of list */
	
	/* nothing can be removed if the list is empty */
	if( listPtr->headNodePtr == NULL )
	{
		fprintf( stderr, "\nERROR: Structure List [%p] is empty!\n", listPtr );
		fprintf( stderr, "Function 'List_removeTail' must not be called for an empty list.\n" );
		AbortProgram;
	}
	
	/* if currentNodePtr points to the tail node it is moved to the previous node */
	if( listPtr->currentNodePtr == listPtr->tailNodePtr )
		listPtr->currentNodePtr = (listPtr->currentNodePtr)->previousNodePtr;
	
	data = (listPtr->tailNodePtr)->data;
	newTailNodePtr = (listPtr->tailNodePtr)->previousNodePtr;
	if( newTailNodePtr != NULL )
		newTailNodePtr->nextNodePtr = NULL;
	free( listPtr->tailNodePtr );
	listPtr->tailNodePtr = newTailNodePtr;

	/* if the last node is removed from the list headNodePtr is set to NULL */
	if( listPtr->tailNodePtr == NULL )
		listPtr->headNodePtr = NULL;

	(listPtr->numberOfNodes)--;
	return data;
}

/** Checks if the structure @ref List is empty.
*
* This function returns @ref BOOLEAN_TRUE if the structure @ref List is empty and @ref BOOLEAN_FALSE
* otherwise.
*
* @param listPtr pointer on the structure @ref List that is checked for emptiness
* @retval int @ref BOOLEAN_TRUE or @ref BOOLEAN_FALSE
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
*/
int List_isEmpty( const List_Ptr listPtr )
{
	if( listPtr->headNodePtr == NULL )
		return BOOLEAN_TRUE;
	else
		return BOOLEAN_FALSE;
}

/** Checks if @c data is contained in the @ref List.
*
* This function returns @ref BOOLEAN_TRUE if @c data is contained in the @ref List and @ref BOOLEAN_FALSE
* otherwise.
*
* @param listPtr pointer on the structure @ref List that is checked for the presence of @c data
* @param data void pointer on the actual data
* @retval int @ref BOOLEAN_TRUE or @ref BOOLEAN_FALSE
* @author Uli Fechner
* @version 03/12/2003 - Uli Fechner - initial release
*/
int List_isContained( const List_Ptr listPtr, const void* data )
{
	ListNode_Ptr tempNodePtr;
	tempNodePtr = listPtr->headNodePtr;

	/* we go from node to node */
	while( tempNodePtr != NULL )
	{
		if( ( listPtr->identicalFunction( tempNodePtr->data, data ) ) == BOOLEAN_TRUE )
			return BOOLEAN_TRUE;
		tempNodePtr = tempNodePtr->nextNodePtr;
	}

	/* we did not find data within the list */
	return BOOLEAN_FALSE;
}

/** The @c numberOfNodes of the @ref List is returned.
*
* @param listPtr pointer on the structure @ref List
* @retval int number of nodes in the @ref List
* @author Uli Fechner
* @version 25/11/2003 - Uli Fechner - initial release
*/
int List_getNumberOfNodes( const List_Ptr listPtr )
{
	return listPtr->numberOfNodes;
}

/** Sets the @c name of the @ref List.
*
* @attention
* The provided parameter @c name is copied. @c name is <b>not</b> freed afterwards. This has to be done
* by the caller of this function!
*
* @param listPtr pointer on the structure @ref List that is associated with @c name
* @param name name of the @ref List
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
*/
void List_setName( const List_Ptr listPtr, char* name )
{
	if( name != NULL )
	{
		if( !( listPtr->name = calloc( strlen( name ) + 1, sizeof( char ) ) ) )
			MemoryError( "List_setName", "listPtr->name" );	
		strncpy( listPtr->name, name, strlen( name ) + 1 );
	}
	else
		listPtr->name = NULL;
}

/** The @c name of the @ref List is returned.
*
* @param listPtr pointer on the structure @ref List that contains the value
* @retval char name of the @ref List
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
*/
char* List_getName( const List_Ptr listPtr )
{
	return listPtr->name;
}

/** The value of the member @c uniqueData of the @ref List is returned.
*
* @param listPtr pointer on the structure @ref List that contains the value
* @retval int value of @c uniqueData of the @ref List
* @author Uli Fechner
* @version 03/12/2003 - Uli Fechner - initial release
*/
int List_getUniqueData( const List_Ptr listPtr )
{
	return listPtr->uniqueData;
}
