/** @file
* This file contains several defines that are used throughout the program.
*
* @author Uli Fechner
* @version 13/05/2003 - Uli Fechner - v1.0.0 - initial release
* @version 24/11/2003 - Uli Fechner - v1.0.1 - modifications to comply to the needs of flux
* @version 29/01/2004 - Tina Grabowski - v1.0.2 - added @ref DAYLIGHT_SMILES, @ref DAYLIGHT_SMIRKS
*                       @ref DAYLIGHT_SMARTS
*/

/* 'simple' define */
#ifndef __GENERALDEFINES_H__
#define __GENERALDEFINES_H__

/* just some information about this file */
#define GENERALDEFINES_VERSION	"1.0.2"
#define GENERALDEFINES_DATE	"29-Jan-2003"

/* general defines */

/** If FLUX_DEBUG is set to 1 some debug information is written to the output file.
*
* @author Uli Fechner
* @version 24/11/2003 - Uli Fechner - initial release
*/
#ifndef COUNTSMARTS_DEBUG
#define COUNTSMARTS_DEBUG 0
#endif

/** Simulates the boolean value @c false.
*
* Together with @ref BOOLEAN_TRUE a boolean variable is simulated in C.
* @author Uli Fechner
* @version 13/05/2003 - Uli Fechner - initial release
*/
#ifndef BOOLEAN_FALSE
#define BOOLEAN_FALSE 0
#endif

/** Simulates the boolean value @c true.
*
* Together with @ref BOOLEAN_FALSE a boolean variable is simulated in C.
* @author Uli Fechner
* @version 13/05/2003 - Uli Fechner - initial release
*/
#ifndef BOOLEAN_TRUE
#define BOOLEAN_TRUE 1
#endif

/** If parameter @c daylight_type in @ref SmilesCompound_create is set to DAYLIGHT_SMILES a SMILES is created
*
* @author Tina Grabowski
* @version 29/01/2004 - Tina Grabowski - initial release
*/
#ifndef DAYLIGHT_SMILES
#define DAYLIGHT_SMILES 1
#endif

/** If parameter @c daylight_type in @ref SmilesCompound_create is set to DAYLIGHT_SMIRKS a SMIRKS is created
*
* @author Tina Grabowski
* @version 29/01/2004 - Tina Grabowski - initial release
*/
#ifndef DAYLIGHT_SMIRKS
#define DAYLIGHT_SMIRKS 2
#endif

/** If parameter @c daylight_type in @ref SmilesCompound_create is set to DAYLIGHT_SMARTS a SMARTS is created
*
* @author Tina Grabowski
* @version 29/01/2004 - Tina Grabowski - initial release
*/
#ifndef DAYLIGHT_SMARTS
#define DAYLIGHT_SMARTS 3
#endif

/* define 'functions' */

/** Final part of the abortion sequence of the program.
*
* If a detectable error occurs during runtime the program initiates an abortion sequence. This consists
* of printing a message on standard error that is related to the error and this define to print a final
* message and exit the program with the exit code EXIT_FAILURE. Such detectable errors include wrong
* command line arguments, missing input files, inconsistent input files, index out of bounds errors,
* memory allocation errors and so on.
* @author Uli Fechner
* @version 05/13/2003 - Uli Fechner - initial release
*/
#ifndef AbortProgram
#define AbortProgram fprintf( stderr, "Program aborted.\n\n" ), exit( EXIT_FAILURE )
#endif

/** Indicates a fatal memory error.
*
* If allocation of memory fails at any time during program execution, this define is called. It prints an
* appropriate message on standard error and calls the define @ref AbortProgram.
*
* @param Variable string containing the name of the variable memory allocation failed
* @param Function string containing the name of the function that tried to allocate memory
* @author Uli Fechner
* @version 05/13/2003 - Uli Fechner - initial release
*/
#ifndef MemoryError
#define MemoryError( Variable, Function ) \
	fprintf( stderr, "\n\nERROR: Fatal memory error.\n" ), \
	fprintf( stderr, "Allocation of memory for %s in function %s failed.\n", Variable, Function ), \
	AbortProgram
#endif

/** Indicates a read error of an input file.
*
* If a read error of an input file occurs at any time during program execution, this define is called. It
* prints a message on standard error and calls the define @ref AbortProgram.
*
* @param FileName string containing the name of the file the read error occured
* @author Uli Fechner
* @version 05/13/2003 - Uli Fechner - initial release
*/
#ifndef FileReadError
#define FileReadError( FileName ) \
	fprintf( stderr, "\n\nERROR: File error.\n" ), \
	fprintf( stderr, "\nCould not read from %s!\n", FileName ), \
	AbortProgram
#endif

/** Indicates a write error of an output file.
*
* If a write error of an output file occurs at any time during program execution, this define is called.
* It prints a message on standard error and calls the define @ref AbortProgram.
*
* @param FileName string containing the name of the file the write error occured
* @author Uli Fechner
* @version 05/13/2003 - Uli Fechner - initial release
*/
#ifndef FileWriteError
#define FileWriteError( FileName ) \
	fprintf( stderr, "\n\nERROR: File error.\n" ), \
	fprintf( stderr, "\nCould not write to %s!\n", FileName ), \
	AbortProgram
#endif

/** Indicates a file integrity error of an input file.
*
* If a file integrity error of an input file occurs at any time during program execution, this define is
* called. It prints a message on standard error and calls the define @ref AbortProgram. A file integrity
* error means that an input file has not the same number of columns in all of its rows.
*
* @param FileName string that contains the name of the file the integrity error occures
* @param Row integer that contains the number of the row the different number of columns appeared
* @param Columns integer containing the number of columns that appear in @c Row
* @param ColumnsBefore integer containing the number of columns that rows have before @c Row
* @author Uli Fechner
* @version 05/13/2003 - Uli Fechner - initial release
*/
#ifndef FileIntegrityError
#define FileIntegrityError( FileName, Row, Columns, ColumnsBefore ) \
	fprintf( stderr, "\n\nERROR: File error.\n" ), \
	fprintf( stderr, "\nFile %s does not contain the same number of columns in all rows!\n", FileName ), \
	fprintf( stderr, "Row %d has %d columns, rows before have %d columns.\n", Row, Columns, ColumnsBefore ), \
	AbortProgram
#endif

/** Indicates an incompatibility of command line arguments/options.
*
* The term @c option refers in the following text to arguments and options provided via the command line
* when starting the program. An @c OptionIncompatibility occurs if one options is not allowed to be used
* with another option @b or one option has to be used together with another option that is not provided.
* This is checked in the function parseClp. If such an event takes place this define is called. It
* prints an appropriate message on standard error and calls the define @ref AbortProgram.
*
* @param Option character with the command line abbreviation of the option causing @c OptionIncompatibility
* @param NotOrExclusively string containing either the word @c not or @c exclusively
* @param RegardedOptions string containing the option(s) @c Option is incompatible with
* @author Uli Fechner
* @version 05/13/2003 - Uli Fechner - initial release
*/
#ifndef OptionIncompatibility
#define OptionIncompatibility( Option, NotOrExclusively, RegardedOptions ) \
	fprintf( stderr, "\nERROR: The %s option may %s ", Option, NotOrExclusively ), \
	fprintf( stderr, "be used together with the %s option!\n", RegardedOptions ), \
	fprintf( stderr, "Type 'speedcats -h' for a detailed help text!\n"), \
	AbortProgram
#endif

/** Indicates a missing command line arguments/options that is mandatory.
*
* The term @c option refers in the following text to arguments and options provided via the command line
* when starting the program. An @c MandatoryOption error occurs if a mandatory option is not provided 
* at the command line. This is checked in the function parseClp. If such an event takes place this
* define is called. It prints an appropriate message on standard error and calls the define @ref
* AbortProgram.
*
* @param Option character with the command line abbreviation of the mandatory option
* @author Uli Fechner
* @version 05/13/2003 - Uli Fechner - initial release
*/
#ifndef MandatoryOption
#define MandatoryOption( Option ) \
	fprintf( stderr, "\nERROR: The %s option is mandatory!\n", Option ), \
	fprintf( stderr, "Type 'retroflux -h' for a detailed help text!\n"), \
	AbortProgram
#endif

/** Indicates an index out of bounds error.
*
* If an element of a structure is accessed that is not available this define is called. Such a situation
* might occur, if the index of the element is bigger than the number of elements the structure has. The
* access could be a read or a write operation. An appropriate message is printed on standard error and
* the define @ref AbortProgram is called.
*
* @param StructureName string containing the name of the structure
* @param NumberOfElements integer containing the number of elements of the structure
* @param Index integer containing the index of the element that should be accessed
* @author Uli Fechner
* @version 05/13/2003 - Uli Fechner - initial release
*/
#ifndef IndexOutOfBoundsError
#define IndexOutOfBoundsError( StructureName, NumberOfElements, Index ) \
	fprintf( stderr, "\n\nERROR: Index out of bounds.\n" ), \
	fprintf( stderr, "Structure %s has only %d elements.\n", StructureName, NumberOfElements ), \
	fprintf( stderr, "These are indexed from 0 to %d.\n", NumberOfElements - 1), \
	fprintf( stderr, "There is no element with index %d!\n\n", Index ), \
	AbortProgram
#endif

#endif /* __GENERALDEFINES_H__ */
