/* just some information about this file */
#define STRINGARRAY_VERSION	"1.3.1"
#define STRINGARRAY_DATE	"19-Dec-2003"

/* structures and typedefs */

/** @file
* @anchor StringArray
* Fault-tolerant structure to store one or more strings.
*
* This structure provides a fault-tolerant way to store several strings. Strings are stored in sequential
* order with the function @ref StringArray_addElement. The individual elements are accessed via their
* index @b starting @b from @b zero. Therefore, the index of the element with the highest index number
* is one less than @c numberOfElements.
*
* The structure is created with @ref StringArray_create, destroyed with @ref StringArray_destroy and
* displayed with @ref StringArray_display. The number of elements of a structure is returned by the
* function @ref StringArray_getNumberOfElements, the size of the array @c arrayOfStrings is returned
* by the function @ref StringArray_getSizeOfArray. Strings are added by @ref StringArray_addElement and
* the value of an element is obtained by @ref StringArray_getElement. Elements that are identical to
* a given string can be removed with @ref StringArray_removeElement.
*
* If an element is accessed via @ref StringArray_getElement and the index is higher than @c
* numberOfElements - 1 (number of elements of this structure) the define @ref IndexOutOfBoundsError is
* called and the program aborts.
*
* The size of @c arrayOfStrings can grow during the lifetime of a structure @ref StringArray. When a
* structure @ref StringArray is created with @ref StringArray_create the initial size of @c
* arrayOfStrings is set to @ref STRINGARRAY_STARTSIZEOFARRAY. @ref StringArray_addElement checks if @c
* numberOfElements exceeds @c sizeOfArray and, if necessary, resizes the string array automatically via
* @ref StringArray_extend. As there is currently no function to remove elements @c arrayOfStrings can
* not shrink during the lifetime of an instance of @ref StringArray!
*
* @attention
* Function @ref StringArray_addElement just links the given char* to @c arrayOfStrings. The char* is not
* copied and therefore must not be freed after the assignment! If @ref StringArray_destroy is called the
* memory of all elements in @c arrayOfStrings is freed automatically.
*
* <B>Description of the members:</B><BR><BR>
* @c arrayOfStrings - array of char*<BR>
* @c numberOfElements - number of elements in the structure<BR>
* @c sizeOfArray - size of the array @c arrayOfStrings<BR>
* @author Uli Fechner
* @version 01/09/2003 - v1.0.0 - Uli Fechner - initial release
* @version 02/09/2003 - v1.1.0 - Uli Fechner - major change: the size of the string array (@c arrayOfStrings) can
* now grow during the lifetime of a structure @ref StringArray; this is handled automatically by @ref
* StringArray_addElement if @c numberOfElements exceeds @c sizeOfArray; added @ref
* STRINGARRAY_STARTSIZEOFARRAY, @ref StringArray_createWithSpecifiedArraySize and @ref
* StringArray_extend; changes in @ref StringArray_create and StringArray_addElement
* @version 08/09/2003 - v1.1.1 - Uli Fechner - small modification in the output format of @ref
* StringArray_display
* @version 08/09/2003 - v1.2.0 - Uli Fechner - bugfix related to the extension of the array (@ref
* StringArray_addElement and @ref StringArray_extend); modified the headers of both functions
* @version 15/10/2003 - v1.2.1 - Uli Fechner - changed the define @ref STRINGARRAY_STARTSIZEOFARRAY
* @version 10/12/2003 - v1.2.2 - Uli Fechner - added function @ref StringArray_concatenate
* @version 17/12/2003 - v1.3.0 - Uli Fechner - added function @ref StringArray_copy and @ref
* StringArray_mapConcatenate
* @version 19/12/2003 - v1.3.1 - Uli Fechner - added function @ref StringArray_removeElement
* @@code
*/
typedef struct
{
	char** arrayOfStrings;
	int numberOfElements;
	int sizeOfArray;
} StringArray;
/** @endcode */


/** A pointer to structure @ref StringArray is assigned the name @c StringArrayPtr. */
typedef StringArray* StringArray_Ptr;

/* defines */

/** This value sets the size of @c arrayOfStrings if a new structure @ref StringArray is created.
*
* @author Uli Fechner
* @version 02/09/2003 - Uli Fechner - initial release
*/
#ifndef STRINGARRAY_STARTSIZEOFARRAY
#define STRINGARRAY_STARTSIZEOFARRAY 50
#endif

/* function prototypes */

StringArray_Ptr StringArray_create( void );

StringArray_Ptr StringArray_createWithSpecifiedArraySize( const int sizeOfArray );

void StringArray_destroy( StringArray_Ptr sAPtr );

void StringArray_extend( const StringArray_Ptr sAPtr );

void StringArray_display( StringArray_Ptr sAPtr, FILE* outputStream );

StringArray_Ptr StringArray_copy( const StringArray_Ptr sAPtr );

void StringArray_concatenate( const StringArray_Ptr sAPtrTarget, const StringArray_Ptr sAPtrSource );

void StringArray_mapConcatenate( const StringArray_Ptr sAPtr, const char* source );

int StringArray_getNumberOfElements( const StringArray_Ptr sAPtr );

int StringArray_getSizeOfArray( const StringArray_Ptr sAPtr );

void StringArray_addElement( const StringArray_Ptr sAPtr, char* string );

char* StringArray_getElement( const StringArray_Ptr sAPtr, const int index );

int StringArray_removeElement( StringArray_Ptr sAPtr, const char* string );

/* functions */

/** Creates a @ref StringArray structure.
*
* A @ref StringArray structure is created. The size of the array @c arrayOfStrings of the structure is
* @ref STRINGARRAY_STARTSIZEOFARRAY and is dynamically adjusted during the lifetime of the structure.
*
* @attention
* This function is only a wrapper for @ref StringArray_createWithSpecifiedArraySize
*
* @retval StringArray_Ptr pointer on the newly created structure @ref StringArray
* @author Uli Fechner
* @version 01/09/2003 - Uli Fechner - initial release
* @version 02/09/2003 - Uli Fechner - major change: now, this function is a wrapper for @ref
*                     StringArray_createWithSpecifiedArraySize; therefore, all instructions moved to
*                     @ref StringArray_createWithSpecifiedArraySize
*/
StringArray_Ptr StringArray_create( void )
{
	return StringArray_createWithSpecifiedArraySize( STRINGARRAY_STARTSIZEOFARRAY );
}

/** Creates a @ref StringArray structure.
*
* A @ref StringArray structure is created. The size of the array @c arrayOfStrings of the structure is
* given as an argument.
*
* @param sizeOfArray the size of the array @c arrayOfStrings of the newly created structure
* @retval StringArray_Ptr pointer on the newly created structure @ref StringArray
* @author Uli Fechner
* @version 02/09/2003 - Uli Fechner - initial release
*/
StringArray_Ptr StringArray_createWithSpecifiedArraySize( const int sizeOfArray )
{
	StringArray_Ptr sAPtr;
	
	if( !( sAPtr = calloc( 1, sizeof( StringArray ) ) ) )
		MemoryError( "sAPtr", "StringArray_createWithSpecifiedArraySize" );
	
	sAPtr->sizeOfArray = sizeOfArray;
	sAPtr->numberOfElements = 0;
	
	if( sizeOfArray == 0 )
		sAPtr->arrayOfStrings = NULL;
	
	else
	{
		if( !( sAPtr->arrayOfStrings = calloc( sizeOfArray, sizeof( char* ) ) ) )
			MemoryError( "sAPtr->arrayOfStrings", "StringArray_createWithSpecifiedArraySize" );
	}

	return sAPtr;
}

/** Destroys a structure @ref StringArray.
*
* The structure @ref StringArray the pointer @c sAPtr refers to is destroyed. All allocated memory of
* the structure is automatically freed; this includes the memory of the char* of @c arrayOfStrings.
*
* @param sAPtr pointer on the structure @ref StringArray that should be destroyed
* @author Uli Fechner
* @version 01/09/2003 - Uli Fechner - initial release
*/
void StringArray_destroy( StringArray_Ptr sAPtr )
{
	int counter;
	
	if( sAPtr != NULL )
	{
		if( sAPtr->arrayOfStrings != NULL )
		{
			for( counter = 0; counter < sAPtr->numberOfElements; counter++ )
			{
				if( (sAPtr->arrayOfStrings)[counter] != NULL )
					free ( (sAPtr->arrayOfStrings)[counter] );
			}
			free( sAPtr->arrayOfStrings );
		}
		free( sAPtr );
	}
	else
	{
		fprintf( stderr, "\n\nERROR: Function 'StringArray_destroy' can not destroy a structure\n" );
		fprintf( stderr, "'StringArray' that has not been created before!\n" );
		AbortProgram;
	}
}


/** Extends the array size of a structure @ref StringArray.
*
* The array size of the structure @ref StringArray is enlarged (more exactly: its size is doubled). This
* function is automatically called by @ref StringArray_addElement if @c numberOfElements exceeds @c
* sizeOfArray.
*
* @param sAPtr pointer on the structure @ref StringArray containing the array that is extended
* @author Uli Fechner
* @version 02/09/2003 - Uli Fechner - initial release
* @version 08/09/2003 - Uli Fechner - function rewritten from scratch; function header modified
*/
void StringArray_extend( const StringArray_Ptr sAPtr )
{
	int counter; /* counter to enumerate the elements of arrayOfIntegers */
	char** newArrayOfStrings; /* replaces iAPtr->arrayOfStrings */
	
	/* create new integer array with doubled size compared to iAPtr->arrayOfString */
	if( !( newArrayOfStrings = calloc( 2 * sAPtr->sizeOfArray, sizeof( char* ) ) ) )
		MemoryError( "newArrayOfStrings", "StringArray_extend" );
	
	/* copy information from sAPtr->arrayOfStrings to newArrayOfStrings */
	for( counter = 0; counter < sAPtr->numberOfElements; counter++ )
		newArrayOfStrings[counter] = (sAPtr->arrayOfStrings)[counter];
	sAPtr->sizeOfArray *= 2;
	
	/* destroy sAPtr->arrayOfStrings */
	if( sAPtr->arrayOfStrings != NULL )
		free( sAPtr->arrayOfStrings );
	
	sAPtr->arrayOfStrings = newArrayOfStrings;
}

/** Displays a structure @ref StringArray.
*
* The structure @ref StringArray the pointer @c sAPtr refers to is displayed on the FILE* @c
* outputStream.
*
* @param sAPtr pointer on the structure @ref StringArray that is displayed
* @param outputStream FILE* on the stream the output is sent to
* @author Uli Fechner
* @version 01/09/2003 - Uli Fechner - initial release
* @version 08/09/2003 - Uli Fechner - small modifications in the output format
*/
void StringArray_display( StringArray_Ptr sAPtr, FILE* outputStream )
{
	int counter; /* simple counter variable in for loop */
	
	if( sAPtr == NULL )
	{
		fprintf( outputStream, "The Structure 'StringArray' (pointer: %p) does not exist.\n\n", sAPtr );
		return;
	}

	if( sAPtr->numberOfElements == 0 )
	{
		fprintf( outputStream, "The Structure 'StringArray' (pointer: %p) is empty.\n\n", sAPtr );
		return;
	}

	for( counter = 0; counter < sAPtr->numberOfElements; counter++ )
		fprintf( outputStream, "%s\t", (sAPtr->arrayOfStrings)[counter] );
	//fprintf( outputStream, "\n" );
}

/** A deepcopy of @c sAPtr is returned.
*
* @attention
* Empty strings (NULL) are not copied. The number of elements of the @b copy of @c sAPtr may
* therefore be less than the number of elements of @c sAPtr.
*
* @param sAPtr a deepcopy of the strings of @c sAPtr is returned
* @retval StringArray_Ptr deepcopy of @c sAPtr
* @author Uli Fechner
* @version 17/12/2003 - Uli Fechner - initial release
*/
StringArray_Ptr StringArray_copy( const StringArray_Ptr sAPtr )
{
	int counter;
	char* currentSourceElement;
	char* tempString;
	StringArray_Ptr copyOfSAPtr;
	
	copyOfSAPtr = StringArray_create( );
	
	for( counter = 0; counter < sAPtr->numberOfElements; counter++ )
	{
		currentSourceElement = StringArray_getElement( sAPtr, counter );
		if( currentSourceElement != NULL )
		{
			if( !( tempString = calloc( strlen( currentSourceElement ) + 1, sizeof( char ) ) ) )
				MemoryError( "tempString", "StringArray_copy" );
			strncpy( tempString, currentSourceElement, strlen( currentSourceElement ) + 1 );
			StringArray_addElement( copyOfSAPtr, tempString );
		}
	}
	return copyOfSAPtr;
}

/** A copy of each string in @c sAPtrSource is added to @c sAPtrTarget.
*
* @attention
* Empty strings (NULL) are not copied. The number of elements that are added to @c sAPtrTarget may
* therefore be less than the number of elements that are added to @c sAPtrSource.
*
* @param sAPtrTarget copies of the strings of @c sAPtrSource are added to this structure @ref StringArray
* @param sAPtrSource copies of the strings of this structure @ref StringArray are added to @c sAPtrSource
* @author Uli Fechner
* @version 10/12/2003 - Uli Fechner - initial release
*/
void StringArray_concatenate( const StringArray_Ptr sAPtrTarget, const StringArray_Ptr sAPtrSource )
{
	int counter;
	char* currentSourceElement;
	char* tempString;
	
	for( counter = 0; counter < sAPtrSource->numberOfElements; counter++ )
	{
		currentSourceElement = StringArray_getElement( sAPtrSource, counter );
		if( currentSourceElement != NULL )
		{
			if( !( tempString = calloc( strlen( currentSourceElement ) + 1, sizeof( char ) ) ) )
				MemoryError( "tempString", "StringArray_concatenate" );	
			strncpy( tempString, currentSourceElement, strlen( currentSourceElement ) + 1 );
			StringArray_addElement( sAPtrTarget, tempString );
		}
	}
}

/** @c source is added to each element (char*) of @c sAPtr.
*
* @attention
* Notrhing is added to empty strings (NULL).
*
* @param sAPtr each element of @c sAPtr is modified
* @param source this char* is concatenated to each element of @c sAPtr
* @author Uli Fechner
* @version 17/12/2003 - Uli Fechner - initial release
*/
void StringArray_mapConcatenate( const StringArray_Ptr sAPtr, const char* source )
{
	int counter;
	char* currentElement;
	char* tempTargetString;	

	for( counter = 0; counter < sAPtr->numberOfElements; counter++ )
	{
		currentElement = StringArray_getElement( sAPtr, counter );
		if( currentElement != NULL )
		{
			if( !( tempTargetString = calloc( strlen( currentElement ) + strlen( source ) + 1, \
				sizeof( char ) ) ) )
				MemoryError( "tempTargetString", "StringArray_mapConcatenate" );
			strncpy( tempTargetString, currentElement, strlen( currentElement ) + 1 );
			strncat( tempTargetString, source, strlen( source ) + 1 );
			(sAPtr->arrayOfStrings)[ counter ] = tempTargetString;
			free( currentElement );
		}
	}		
}

/** Returns the number of elements of a structure @ref StringArray.
*
* The number of elements of the structure @ref StringArray the pointer @c sAPtr refers to is returned.
*
* @param sAPtr pointer on the structure @ref StringArray of which the number of elements is returned
* @retval int number of elements of the structure @ref StringArray
* @author Uli Fechner
* @version 01/09/2003 - Uli Fechner - initial release
*/
int StringArray_getNumberOfElements( const StringArray_Ptr sAPtr )
{
	return sAPtr->numberOfElements;
}

/** Returns the size of the array @c arrayOfStrings of a structure @ref StringArray.
*
* The size of the array @c arrayOfStrings of the structure @ref StringArray is returned.
*
* @param sAPtr pointer on the structure @ref StringArray of which the size of the array is returned
* @retval int size of the array @c arrayOfStrings of the structure @ref StringArray
* @author Uli Fechner
* @version 01/09/2003 - Uli Fechner - initial release
*/
int StringArray_getSizeOfArray( const StringArray_Ptr sAPtr )
{
	return sAPtr->sizeOfArray;
}

/** A char* is added to the structure @ref StringArray.
*
* The char* is added to the array @c arrayOfStrings of structure @ref StringArray.
* If the array @c arrayOfStrings is already full (i.e. @c numberOfElements >= @c sizeOfArray)
* it is extended by calling the function @ref StringArray_extend.
*
* @param sAPtr pointer on the structure @ref StringArray the char* is added to
* @param string char* that is added to @ref StringArray
* @author Uli Fechner
* @version 01/09/2003 - Uli Fechner - initial release
* @version 02/09/2003 - Uli Fechner - added the capability to extend @c arrayOfStrings if necessary;
*                     therefore, @b const of @c sAPtr had to be removed
* @version 08/09/2003 - Uli Fechner - fixed a bug related to the function call @ref StringArray_extend;
* function header modified (const added)
*/
void StringArray_addElement( const StringArray_Ptr sAPtr, char* string )
{
	if( sAPtr->numberOfElements >= sAPtr->sizeOfArray )
		StringArray_extend( sAPtr );

	(sAPtr->arrayOfStrings)[(sAPtr->numberOfElements)] = string;
	(sAPtr->numberOfElements)++;
}

/** The char* of the element with index @c index of structure @ref StringArray is returned.
*
* The value of the element with index @c index of the structure @ref StringArray @c sAPtr points on is
* returned. If the index @c index is greater than (@c numberOfElements - 1 ), the program aborts with
* the call of the define @ref IndexOutOfBoundsError.
*
* @param sAPtr pointer on the structure @ref StringArray that contains the char*
* @param index index of the char* that is returned
* @retval char* the char* with index @c index
* @author Uli Fechner
* @version 01/09/2003 - Uli Fechner - initial release
*/
char* StringArray_getElement( const StringArray_Ptr sAPtr, const int index )
{
	if( index < sAPtr->numberOfElements )
		return (sAPtr->arrayOfStrings)[ index ];
	else
		IndexOutOfBoundsError( "StringArray", sAPtr->numberOfElements, index );
}

/** All elements of @c sAPtr that are identical to @c string are removed.
*
* Elements that are identical to @c string are removed from @c sAPtr. The return value is the number
* of times @c string is found in @c sAPtr.
*
* @param sAPtr @c string is removed from the structure @ref StringArray that is referred to by @c sAPtr
* @param string all occurrences of string are removed from @c sAPtr
* @retval int number of times @c string was removed from @c sAPtr
* @author Uli Fechner
* @version 19/12/2003 - Uli Fechner - initial release
*/
int StringArray_removeElement( StringArray_Ptr sAPtr, const char* string )
{
	StringArray_Ptr newSAPtr; /* new StringArray: all elements that equal string are removed */
	int counter; /* simple counter in for loop */
	int stringFound = 0; /* stores the number of times string is found within sAPtr */
	
	for( counter = 0; counter < sAPtr->numberOfElements; counter++ )
	{
		if( strcmp( sAPtr->arrayOfStrings[ counter ], string ) == 0 )
		{
			stringFound++;
			free( sAPtr->arrayOfStrings[ counter ] );
			sAPtr->arrayOfStrings[ counter ] = NULL;
		}
	}
	newSAPtr = StringArray_copy( sAPtr );
	StringArray_destroy( sAPtr );
	sAPtr = newSAPtr;

	return stringFound;
}
